import multiprocessing
from aquests.lib import pathtool, logger

class Logger:
	def __init__ (self, media, path):
		self.media = type (media) is list and media  or [media]		
		self.path = path
		if self.path: 
			pathtool.mkdir (path)			
		self.logger_factory = {}
		self.lock = multiprocessing.Lock ()
		
		self.make_logger ("server", "monthly")
		self.make_logger ("app", "daily")
		self.make_logger ("request", "daily")
		
	def make_logger (self, prefix, freq = "daily"):
		self.lock.acquire ()
		has_prefix = prefix in self.logger_factory
		if has_prefix:
			self.lock.release ()
			raise TypeError("%s is already used" % prefix)
								
		_logger = logger.multi_logger ()
		if self.path and 'file' in self.media:
			_logger.add_logger (logger.rotate_logger (self.path, prefix, freq))
		if 'screen' in self.media:
			_logger.add_logger (logger.screen_logger ())
		
		self.logger_factory [prefix] = _logger
		self.lock.release ()	
	
	def add_screen_logger (self):
		for prefix, _logger in list(self.logger_factory.items ()):
			_logger.add_logger (logger.screen_logger ())
		
	def get (self, prefix):
		return self.logger_factory [prefix]
			
	def trace (self, prefix, ident = ""):
		self.get (prefix).trace (ident)		
		
	def __call__ (self, prefix, msg, log_type = ""):		
		self.get (prefix).log (msg, log_type)
	
	def rotate (self):
		self.lock.acquire ()
		loggers = list(self.logger_factory.values ())
		self.lock.release ()
		
		for mlogger in loggers:
			for logger in mlogger.loggers:
				if hasattr (logger, "rotate"):
					logger.rotate ()
		
	def close (self):
		with self.lock:
			for mlogger in self.logger_factory.values ():
				mlogger.close ()
			self.logger_factory = {}
			


"""
2015. 12. 10
Hans Roh
"""
from .Saddle import Saddle

# Events
app_starting = "app:starting"
app_started = "app:started"
app_restarting = "app:restarting"
app_restarted = "app:restarted"
app_mounted = "app:mounted"
app_unmounting = "app:umounting"

request_failed = "request:failed"
request_success = "request:success"
request_tearing_down = "request:teardown"
request_starting = "request:started"
request_finished = "request:finished"

template_rendering = "template:rendering"
template_rendered = "template:rendered"




import json
import sys
from datetime import date
from xmlrpc import client

def catch (format = 0, exc_info = None):
	if exc_info is None:
		exc_info = sys.exc_info()
	t, v, tb = exc_info
	tbinfo = []
	assert tb # Must have a traceback
	while tb:
		tbinfo.append((
			tb.tb_frame.f_code.co_filename,
			tb.tb_frame.f_code.co_name,
			str(tb.tb_lineno)
			))
		tb = tb.tb_next

	del tb
	file, function, line = tbinfo [-1]
	
	# format 0 - text
	# format 1 - html
	# format 2 - list
	try:
		v = str (v)
	except:
		v = repr (v)
		
	if format == 1:
		if v.find ('\n') != -1:
			v = v.replace ("\r", "").replace ("\n", "<br>")
		buf = ["<h3>%s</h3><h4>%s</h4>" % (t.__name__.replace (">", "&gt;").replace ("<", "&lt;"), v)]
		buf.append ("<b>at %s at line %s, %s</b>" % (file, line, function == "?" and "__main__" or "function " + function))
		buf.append ("<ul type='square'>")
		buf += ["<li>%s <span class='f'>%s</span> <span class='n'><b>%s</b></font></li>" % x for x in tbinfo]
		buf.append ("</ul>")		
		return "\n".join (buf)

	buf = []
	buf.append ("%s %s" % (t, v))
	buf.append ("in file %s at line %s, %s" % (file, line, function == "?" and "__main__" or "function " + function))
	buf += ["%s %s %s" % x for x in tbinfo]
	if format == 0:
		return "\n".join (buf)
	return buf	

	
class DateEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, date):
			return str(obj)			
		return json.JSONEncoder.default(self, obj)
		
			
class API:
	def __init__ (self, request, data = None):
		self.request = request # for response typing 
		self.data = data or {}
		self.content_type = self.set_content_type ()
		
	def __setitem__ (self, k, v):
		self.set (k, v)
	
	def __getitem__ (self, k, v = None):
		return self.get (k, v)
	
	def set (self, k, v):
		self.data [k] = v
	
	def get (self, k, v = None):
		return self.data.get (k, )
	
	def get_content_type (self):
		return self.content_type
		
	def set_content_type (self):
		content_type = self.request.get_header ("accept", 'application/json')
		if not content_type.startswith ("text/xml"):			
			content_type = 'application/json'
		return content_type	
		
	def encode (self, charset):
		return self.to_string ().encode (charset)
	
	def __str__ (self):
		return self.to_string ()
		
	def to_string (self):
		if self.content_type.startswith ("text/xml"):
			return client.dumps ((self.data,))
		return json.dumps (self.data, cls = DateEncoder)
			
	def traceback (self, message = 'exception occured', code = 20000, debug = 'see traceback', more_info = None):
		self.error (message, code, debug, more_info, sys.exc_info ())
		
	def error (self, message, code = 20000, debug = None, more_info = None, exc_info = None):
		self.data = {}
		self.data ['code'] = code
		self.data ['message'] = message
		if debug:
			self.data ['debug'] = debug
		if more_info:
			self.data ['more_info'] = more_info
		if exc_info:
			self.data ["traceback"] = catch (2, exc_info)




# pytest framework ---------------------------------------------
import requests
from aquests.lib import siesta
from aquests.lib.pmaster import Puppet, processutil
import subprocess
import time
import sys
import os

class launch:
    def __init__ (self, script, port = 5000, ssl = False, silent = True):
        self.__script = script
        endpoint = "http{}://127.0.0.1".format (ssl and "s" or "")
        if port:
            endpoint += ":{}".format (port)
        self.__endpoint = endpoint
        self.__silent = silent
        self.__servicing = False
        self.__p = Puppet (communicate = False)        
        self.__requests = Requests (endpoint)
        self.__api = siesta.API (endpoint)
        self.__closed = True
        self.__start ()
    
    def __enter__ (self):
        return self
        
    def __exit__ (self, type, value, tb):
        self._close ()
    
    def __del__ (self):
        self._close ()
         
    def __getattr__ (self, attr):
        if attr in ("get", "post", "put", "patch", "delete", "head", "options"):
            return getattr (self.__requests, attr)
        else:
            return getattr (self.__api, attr)
        
    def __start (self):
        if self.__silent:
            if self.__is_running ():
                self.__servicing = True
            else:    
                self.__p.start ([sys.executable, self.__script, "-d"])
                self.__wait_until ("running")
        else:
            self.__p.start ([sys.executable, self.__script])
            time.sleep (3)
        self.__closed = False    
    
    def __is_running (self):
        command = "{} {} status".format (sys.executable, self.__script)
        res = subprocess.run (command, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)    
        out, err = res.stdout.decode ("utf8"), res.stderr.decode ("utf8")
        return out.find ("running") != -1

    def __wait_until (self, status, timeout = 10):
        for i in range (timeout):
            command = "{} {} status".format (sys.executable, self.__script)
            res = subprocess.run (command, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)    
            out, err = res.stdout.decode ("utf8"), res.stderr.decode ("utf8")
            if out.find (status) != -1:
                return
            time.sleep (1)
    
    def _close (self):
        if self.__closed:
            return        
        if self.__silent and not self.__servicing:
            self.__p.start ([sys.executable, self.__script, "stop"])
            self.__wait_until ("stopped")
        else:
            self.__p.kill ()
            time.sleep (3)
        self.__closed = True            


class Requests:
    def __init__ (self, endpoint):
        self.endpoint = endpoint
        self.s = requests.Session ()
    
    def resolve (self, url):
        if url.startswith ("http://") or url.startswith ("https://"):
            return url
        else:
            return self.endpoint + url 
        
    def get (self, url, *args, **karg):
        return self.s.get (self.resolve (url), *args, **karg)
        
    def post (self, url, *args, **karg):
        return self.s.post (self.resolve (url), *args, **karg)
    
    def put (self, url, *args, **karg):
        return self.s.put (self.resolve (url), *args, **karg)
    
    def patch (self, url, *args, **karg):
        return self.s.patch (self.resolve (url), *args, **karg)
    
    def delete (self, url, *args, **karg):
        return self.s.delete (self.resolve (url), *args, **karg)
    
    def head (self, url, *args, **karg):
        return self.s.head (self.resolve (url), *args, **karg)
                
    def options (self, url, *args, **karg):
        return self.s.options (self.resolve (url), *args, **karg)
          


from . import xmlrpc_executor
try:
	import jsonrpclib
except ImportError:
	pass

class Executor (xmlrpc_executor.Executor):
	
	def __call__ (self):
		request = self.env ["skitai.was"].request		
		data = self.env ["wsgi.input"].read ()
		args = jsonrpclib.loads (data)

		is_multicall = False		
		jsonrpc = "2.0"		
		if isinstance (args, dict):
			thunks = [(args ["method"], args.get ("params", []), args ["id"], args ["jsonrpc"])]						
		else:
			is_multicall = True
			thunks = []
			for each in args:
				thunks.append ((each ["method"], each.get ("params", []), each ['id'], each ['jsonrpc']))
				
		self.build_was ()				
		results = []		
		for _method, _args, _rpcid, _jsonrpc in thunks:
			path_info = self.env ["PATH_INFO"] = "/" + _method.replace (".", "/")						
			current_app, thing, param, respcode = self.find_method (request, path_info, is_multicall is False)
			if respcode:				
				results.append (jsonrpclib.dumps (jsonrpclib.Fault (-32000, self.get_http_error_message (respcode), rpcid = _rpcid)))
				continue
				
			self.was.subapp = current_app
			try:
				result = self.chained_exec (thing, _args, {})
			except:			   
				results.append (jsonrpclib.dumps (jsonrpclib.Fault (-32000, self.get_traceback (), rpcid = _rpcid)))
			else:
				results.append (jsonrpclib.dumps (result, methodresponse = True, rpcid = _rpcid, version = _jsonrpc))
			del self.was.subapp
		
		self.was.response ["Content-Type"] = "application/json-rpc"
		self.commit ()
		del self.was.env   
		
		return is_multicall and  ("[" + ",".join (results) + "]") or results [0]
		


import pstats
import os
from aquests.lib.termcolor import tc

def main ():
    if not os.path.isfile ('profile.out'):
        print (tc.error ("error:"))
        print (" - profile.out file not found")
        print (" - for creating this file, run your app with --skitai-profile")
        print ()
        return
        
    p = pstats.Stats ('profile.out')
    p.strip_dirs().sort_stats("ncalls").print_stats(100)
    
